from math import sqrt


def reverse_int(num: (int, str)) -> int:
    """
    returns a new reversed int from the given int or string with int
    :param num the number to reverse

    Example:
        >>> reverse_int(5678)\n
        >>> 8765

    """
    return int(str(num)[::-1])


def persistence(num: int) -> int:
    """
    returns the persistence of a number\
    :param num: the number to check
    :param num: the number to calculate the persistence of

    Example:
        >>> persistence(1234)
        >>> 2

    """
    string_num = str(num)
    count: int = 0
    while len(string_num) > 1:
        result = 1
        for i in string_num:
            result *= int(i)
        string_num = str(result)
        count += 1
    return count


def is_square(n: int) -> bool:
    """
    returns True if n is a perfect square, False otherwise
    :param n: the number to check

    Example:
        >>> is_square(4)
        >>> True
        >>> is_square(5)
        >>> False
    """
    return int(sqrt(n)) ** 2 == n


def is_hexagonal(n: int) -> bool:
    """
    returns True if n is a hexagonal number, False otherwise
    :param n: the number to check

    Example:
        >>> is_hexagonal(6)
        >>> True
        >>> is_hexagonal(5)
        >>> False
    """
    return n > -1 and sqrt(8 * n + 1) % 4 == 3


def is_triangular(n: int) -> bool:
    """
    returns True if n is a triangular number, False otherwise
    :param n: the number to check

    Example:
        >>> is_triangular(6)
        >>> True
        >>> is_triangular(5)
        >>> False
    """
    return n > -1 and sqrt(8 * n + 1) % 1 == 0


def is_pentagonal(n: int) -> bool:
    """
    returns True if n is a pentagonal number, False otherwise
    :param n: the number to check

    Example:
        >>> is_pentagonal(5)
        >>> True
        >>> is_pentagonal(6)
        >>> False
    """
    return n > -1 and sqrt(24 * n + 1) % 6 == 5


def is_prime(n: int) -> bool:
    """
    returns True if n is a prime number, False otherwise
    :param n: the number to check

    Example:
        >>> is_prime(5)
        >>> True
        >>> is_prime(6)
        >>> False
    """
    if n < 2:
        return False
    for i in range(2, int(sqrt(n)) + 1):
        if n % i == 0:
            return False
    return True


def lerp(a: float, b: float, t: float) -> float:
    """
    returns a linear interpolation between a and b at t
    :param a: the first value
    :param b: the second value
    :param t: the interpolation value

    Example:
        >>> lerp(0, 1, 0.5)
        >>> 0.5

    """
    return a + t * (b - a)


def clamp(value: float, n_min: float, n_max: float) -> float:
    """
    returns a clamped value between min and max
    :param value: the value to clamp
    :param n_min: the minimum value
    :param n_max: the maximum value

    Example:
        >>> clamp(12, 0, 10)
        >>> 10
    """
    return max(min(value, n_max), n_min)


def clamp01(value: float) -> float:
    """
    returns a clamped value between 0 and 1
    :param value: the value to clamp

    Example:
        >>> clamp01(12)
        >>> 1
    """
    return clamp(value, 0, 1)


def prime_factors(num: int) -> list:
    """
    returns a list of prime factors of a number
    :param num: the number to get the prime factors of
    :return: list of prime factors

    Example:
        >>> prime_factors(12)
        >>> [2, 2, 3]

    """
    factors = []
    i = 2
    while i * i <= num:
        if num % i:
            i += 1
        else:
            num //= i
            factors.append(i)
    if num > 1:
        factors.append(num)
    return factors



def remove_duplicates(list1: list) -> list:
    """
    Removes duplicates from a list
    :param list1: list to remove duplicates from
    :return: list with duplicates removed

        Example:
            >>> remove_duplicates([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4])
            >>> [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    return list(dict.fromkeys(list1))


def remove_duplicates_sort(list1: list) -> list:
    """
    Removes duplicates from a list with sorted elements also works with strings and numbers
    :param list1: list to remove duplicates from
    :return: list with duplicates removed

        Example:
            >>> remove_duplicates_sort([3, 1, 2 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4])
            >>> [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    return list(set(list1))


def sort_list(list1: list, from_to_types: (tuple, list), sort_elements: bool = False) -> list:
    """
    Sorts a list for multiple types also sorts by elements if sort_elements is True

    warning: if there is a list in the list, that list will be sorted by type 'list' if type not specified it will be
    in the last position

    :param sort_elements: whether to sort the elements of the list or just sort by type (sort by type is default)
    :param list1: list to sort
    :param from_to_types: tuple or list with the type of the list elements to sort by
    :return: sorted list

        Example:
            >>> test_list = [2, 3, 2, 1, "string1", "string2", "string3",  3.3, 2.2]
            >>> sort_list(test_list, (str, int, float))
            >>> ['string1', 'string2', 'string3', 2, 3, 2, 1, 3.3, 2.2]
            >>> sort_list(test_list, (str, int, float), sort_elements=True)
            >>> ['string1', 'string2', 'string3', 1, 2, 2, 3, 2.2, 3.3]
    """
    end_list: list = []
    return_list = []
    curr_idx = 0
    types_separated_list = []
    if sort_elements:
        for types in from_to_types:
            types_item = []
            for item in list1:
                if types == type(item):
                    types_item.append(item)
            types_separated_list.append(types_item)
        sorted_list = []
        for item in types_separated_list:
            sorted_list.append(sorted(item))
        for item in sorted_list:
            for element in item:
                return_list.append(element)
    else:
        for types in from_to_types:
            for item in list1:
                if types == type(item):
                    return_list.insert(curr_idx, item)
                    curr_idx += 1
    for i in list1:
        if type(i) not in from_to_types:
            end_list.append(i)

    return return_list + end_list


def subdivide_by_type(list1: list, from_to_types: (tuple, list)) -> list:
    """
    Subdivides a list by types

    :param list1: list to subdivide
    :param from_to_types: tuple or list with the type of the list elements to subdivide by
    :return: list with the subdivided list

        Example:
            >>> test_list = [2, 3, 2, 1, "string1", "string2", "string3",  3.3, 2.2]
            >>> subdivide_by_type(test_list, (str, int, float))
            >>> [['string1', 'string2', 'string3'], [2, 3, 2, 1], [3.3, 2.2]]
    """
    return_list = []
    for types in from_to_types:
        types_item = []
        for item in list1:
            if types == type(item):
                types_item.append(item)
        return_list.append(types_item)
    return return_list






#
# MIT License
# Copyright (c) 2021 FUSEN
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from string import ascii_letters
from random import sample, randint

_MORSE_DECODE_DICT = {"-.-.--": "!", ".-..-.": '"', "...-..-": "$", ".-...": "&", ".----.": "\\", "-.--.": "(",
                      "-.--.-": ")", ".-.-.": "+", "--..--": ",", "-....-": "-", ".-.-.-": ".", "-..-.": "/",
                      "-----": "0", ".----": "1", "..---": "2", "...--": "3", "....-": "4", ".....": "5", "-....": "6",
                      "--...": "7", "---..": "8", "----.": "9", "---...": ":", "-.-.-.": ";", "-...-": "=",
                      "..--..": "?", ".--.-.": "@", ".-": "A", "-...": "B", "-.-.": "C", "-..": "D", ".": "E",
                      "..-.": "F", "--.": "G", "....": "H", "..": "I", ".---": "J", "-.-": "K", ".-..": "L", "--": "M",
                      "-.": "N", "---": "O", ".--.": "P", "--.-": "Q", ".-.": "R", "...": "S", "-": "T", "..-": "U",
                      "...-": "V", ".--": "W", "-..-": "X", "-.--": "Y", "--..": "Z", "..--.-": "_",
                      "...---...": "SOS"}
_MORSE_ENCODE_DICT = {'A': '.-', 'B': '-...', 'C': '-.-.', 'D': '-..', 'E': '.', 'F': '..-.', 'G': '--.', 'H': '....',
                      'I': '..', 'J': '.---', 'K': '-.-', 'L': '.-..', 'M': '--', 'N': '-.', 'O': '---', 'P': '.--.',
                      'Q': '--.-', 'R': '.-.', 'S': '...', 'T': '-', 'U': '..-', 'V': '...-', 'W': '.--', 'X': '-..-',
                      'Y': '-.--', 'Z': '--..', '1': '.----', '2': '..---', '3': '...--', '4': '....-', '5': '.....',
                      '6': '-....', '7': '--...', '8': '---..', '9': '----.', '0': '-----', ', ': '--..--',
                      '.': '.-.-.-', '?': '..--..', '/': '-..-.', '-': '-....-', '(': '-.--.', ')': '-.--.-'}


def to_binary(string: str) -> str:
    """
    Convert String to numerical binary numbers
    returns numerical binary string
    :param string: String to convert

    Example:
        >>> to_binary("test")
        >>> "1110100 1100101 1110011 1110100"
    """
    bin_conv = []
    for c in string:
        ascii_val = ord(c)
        binary_val = bin(ascii_val)
        bin_conv.append(binary_val[2:])

    return ' '.join(bin_conv)


def bin_to_text(string: str) -> str:
    """
    Convert Binary Numerical Values back to string
    returns utf-8 String
    :param string: Binary String
    :return: str

    Example:
        >>> bin_to_text("1110100 1100101 1110011 1110100")
        >>> 'test'
    """
    return "".join([chr(int(binary, 2)) for binary in string.split(" ")])


def reverse_string(string: str) -> str:
    """
    Returns a new reversed string
    :param string: String to reverse
    :return: str

    Example:
        >>> reverse_string("test")
        >>> "tset"


    """
    return string[::-1]


def encodeMorse(sequence: str) -> str:
    """
       Encodes the given string to Morse Code

       Morse code is case-insensitive and traditonally Capital Letters are used
       Lower cases are converted to Upper cases

       Every Character code is seperated by a single space between them

       A space in the sequence is translated as '   ' three spaces
    :param sequence: String to encode
    :return: str

    Example:
        >>> encodeMorse("HELLO WORLD")
        >>> ".... . .-.. .-.. ---   .-- --- .-. .-.. -.."
    """
    encoded = ''
    for letter in sequence.upper():
        if letter != ' ':
            encoded += _MORSE_ENCODE_DICT[letter] + ' '
        else:
            encoded += '  '

    return encoded.strip()


def decodeMorse(sequence: str) -> str:
    """
    Fully Decodes to Morse codes encoded using encodeMorse() to Upper Case Alphabets
    //Morse codes encoded using external encoders may raise error//
    :param sequence: String to decode
    :return: str

    Example:
        >>> decodeMorse(".... . .-.. .-.. ---   .-- --- .-. .-.. -..")
        >>> "HELLO WORLD"

    """
    return ' '.join(
        ''.join(_MORSE_DECODE_DICT[letter] for letter in word.split(' ')) for word in sequence.strip().split('   '))


def encodeCaesar(sequence: str, shift: int = 1) -> str:
    """
    Encodes the given string to Caesar Cipher
    :param sequence: String to encode
    :param shift: Shift value

    Example:
        >>> encodeCaesar("HELLO WORLD", 1)
        >>> "IFMMP XPSME"
    """
    return ''.join(chr(((ord(c) - 65 + shift) % 26) + 65) if 65 <= ord(c) <= 90 else c for c in sequence)


def decodeCaesar(sequence: str, shift: int = 1) -> str:
    """
    Decodes the given string to Caesar Cipher
    :param sequence: String to decode
    :param shift: Shift value

    Example:
        >>> decodeCaesar("IFMMP XPSME", 1)
        >>> "HELLO WORLD"
    """
    return ''.join(chr(((ord(c) - 65 - shift) % 26) + 65) if 65 <= ord(c) <= 90 else c for c in sequence)


def split_str(string: str, maxsplit: int = 1) -> list:
    """
    Splits characters of a String into List of Strings
    :param string: String to split
    :param maxsplit: It is the number of skips in character before splitting, DEFAULT = 1
    :return: Returns the Array containing elements of characters splitted from the String

    Example:
        >>> split_str("HELLO WORLD", 2)
        >>> ['He', 'll', 'o ', 'Wo', 'rl']

    """
    txt = ""
    str_list = []
    for i in string:
        txt += i
        if len(txt) == maxsplit:
            str_list.append(txt)
            txt = ''
    return str_list


def is_pangram(string: str) -> bool:
    """
    Checks if the given string is a pangram or not
    :param string: String to check
    :return: bool

    Example:
        >>> is_pangram("The quick, brown fox jumps over the lazy dog!")
        >>> True

    """
    count = 0
    for character in range(97, 123):
        if chr(character) in string or chr(character).upper() in string:
            count += 1
    if count == 26: return True
    return False


def is_palindrome(string: str) -> bool:
    """
    Checks if the given string is a palindrome or not
    :param string: String to check
    :return: bool

    Example:
        >>> is_palindrome("ABA")\n
        >>> True

    """
    if string == string[::-1]:
        return True
    return False


def is_anagram(str1: str, str2: str) -> bool:
    """
    Checks if the given string is an anagram of another string
    :param str1: String to check
    :param str2: String to check
    :return: bool

    Example:
        >>> is_anagram("waterbottle", "erbottlewat")
        >>> True

    """
    if len(str1) != len(str2):
        return False
    return sorted(str1) == sorted(str2)


def is_rotation(str1: str, str2: str) -> bool:
    """
    Checks if the given string is a rotation of another string
    :param str1: String to check
    :param str2: String to check
    :return: bool

    Example:
        >>> is_rotation("waterbottle", "erbottlewat")
        >>> True

    """
    if len(str1) != len(str2):
        return False
    return str2 in (str1 + str1)


def make_n_gram(string: str, n: int) -> list:
    """
    Creates n-grams from the given string

    :param string: Specify a input String
    :param n: Specify the n-gram size
    :return: Returns the Array containing n-grams

    Example:
        >>> make_n_gram("Hello world", 2)
        >>> ['He', 'el', 'll', 'lo', 'o ', ' w', 'wo', 'or', 'rl', 'ld']

    """
    n_grams = []
    for i in range(len(string) - n + 1):
        n_grams.append(string[i:i + n])
    return n_grams


def randomize_string(string: str) -> str:
    """
    Randomizes the given string
    :param string: String to randomize
    :return: str

    Example:
        >>> randomize_string("Hello World")
        >>> 'rldloHW'

    """
    return ''.join(sample(string, len(string)))


def get_random_color() -> str:
    """
    Returns a random hex color
    :return: str


    Example:
        >>> get_random_color()
        >>> '#ff0000'

    """
    return '#{:06x}'.format(randint(0, 0xFFFFFF))


def get_random_string(length: int) -> str:
    """
    Returns a random string of given length
    :param length: Length of the string
    :return: str

    Example:
        >>> get_random_string(10)
        >>> 'qwertzuiop'

    """
    return ''.join(sample(ascii_letters, length))


def count_vowels(string: str) -> int:
    """
    Counts the number of vowels in the given string
    :param string: String to count
    :return: int

    Example:
        >>> count_vowels("Hello World")
        >>> 3

    """
    return sum(1 for char in string if char.lower() in 'aeiou')


def count_consonants(string: str) -> int:
    """
    Counts the number of consonants in the given string
    :param string: String to count
    :return: int

    Example:
        >>> count_consonants("Hello World")
        >>> 7

    """
    return sum(1 for char in string if char.lower() in 'bcdfghjklmnpqrstvwxyz')


def count_lines(string: str) -> int:
    """
    Counts the number of lines in the given string
    :param string: String to count
    :return: int

    Example:
        >>> count_lines("Hello World")
        >>> 1

    """
    return len(string.splitlines())





# file_tools.py has useful functions for working with files
import os


def get_size(file_path: str, size_in="") -> str:
    """
    Returns file size in gb/mb/kb/bytes
    :param  file_path: path to the file
    :param  size_in: gb/mb/kb/bytes (if not set default would be its original size)
    :return: file size in gb/mb/kb/bytes

        Example:
            >>> get_size("test.txt")
            >>> "8.98 KB"
            >>> get_size("test.txt", size_in="bytes")
            >>> "9200 bytes"
    """
    if size_in == "":
        size = os.path.getsize(file_path)
        if size > 1024 ** 3:
            size = size / 1024 ** 3
            return f"{size:.2f} GB"
        elif size > 1024 ** 2:
            size = size / 1024 ** 2
            return f"{size:.2f} MB"
        elif size > 1024:
            size = size / 1024
            return f"{size:.2f} KB"
        else:
            return f"{size} bytes"
    else:
        size = os.path.getsize(file_path)
        if size_in == "gb":
            size = size / 1024 ** 3
            return f"{size:.2f} GB"
        elif size_in == "mb":
            size = size / 1024 ** 2
            return f"{size:.2f} MB"
        elif size_in == "kb":
            size = size / 1024
            return f"{size:.2f} KB"
        elif size_in == "bytes":
            return f"{size} bytes"
        else:
            raise ValueError(f"{size_in} is not a valid size type")


def find_in_files(folder_path: str, search_string: str, file_types: list = None, case_sensitive: bool = False) -> list:
    """
    Returns a list of files that contains the given string and the line number and also the entire line that contains.
    :param folder_path: path to the folder
    :param search_string: string to search for
    :param file_types file_types to target (optional)
    :param case_sensitive: if True will search case-sensitive
    :return: list of files that contains the given string and the line number and also the entire line that contains

        Example:
            >>> find_in_files("/home/user/Desktop", "test")
            >>> [["test.txt", 1, "test"], ["test2.txt", 2, "test"]]

    """
    if file_types is None:
        file_types = []
    files = []
    if not file_types:
        for file in os.listdir(folder_path):
            with open(os.path.join(folder_path, file), "r") as f:
                for i, line in enumerate(f):
                    if case_sensitive:
                        if search_string in line:
                            files.append([file, i + 1, line])
                    else:
                        if search_string.lower() in line.lower():
                            files.append([file, i + 1, line])
    else:
        for file_type in file_types:
            for file in os.listdir(folder_path):
                if file.endswith(file_type):
                    with open(os.path.join(folder_path, file), "r") as f:
                        for i, line in enumerate(f):
                            if case_sensitive:
                                if search_string in line:
                                    files.append([file, i + 1, line])
                            else:
                                if search_string.lower() in line.lower():
                                    files.append([file, i + 1, line])
    return files






from math import floor, ceil, trunc, acos


# Vector2
class Vec2:
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def __int__(self, other: "Vec2"):
        self.x = other.x
        self.y = other.y

    def distance(self, other: "Vec2"):
        """
        Returns the distance between two points.
        :param other: The other point.
        :return: The distance.
        """
        return abs(self - other)

    def scale(self, other: "Vec2"):
        """
        Returns the scaled vector.
        :param other: The scaling vector.
        :return: The scaled vector.
        """
        return Vec2(self.x * other.x, self.y * other.y)

    def reflect(self, normal):
        """
        Returns the reflected vector.
        :param normal: The normal vector.
        :return: The reflected vector.
        """
        return self - 2 * self.dot(normal) * normal

    def angle_between(self, other):
        """
        Returns the angle between two vectors.
        :param other: The other vector.
        :return: The angle.
        """
        return acos(self.dot(other) / (self.length() * other.length()))

    @property
    def normalize(self):
        """
        Returns a normalized vector.
        :return: The normalized vector.
        """
        return self / self.length()

    def dot(self, other):
        """
        Returns the dot product of two vectors.
        :param other: The other vector.
        :return: The dot product.
        """
        return self.x * other.x + self.y * other.y

    def length(self):
        """
        Returns the length of the vector.
        :return: The magnitude.
        """
        return (self.x ** 2 + self.y ** 2) ** 0.5

    def length_sqr(self):
        """
        Returns the squared length of the vector.
        :return: The squared magnitude.
        """
        return self.x ** 2 + self.y ** 2

    @staticmethod
    def angle(a, b):
        """
        Returns the angle between two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The angle.
        """
        return acos(a.dot(b) / (a.length() * b.length()))

    @staticmethod
    def zero():
        """
        Returns a zero vector.
        :return: The zero vector.
        """
        return Vec2(0, 0)

    @staticmethod
    def cross(a, b):
        """
        Returns the cross product of two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The cross product.
        """
        return a.x * b.y - a.y * b.x

    @staticmethod
    def up():
        """
        Returns a vector pointing up.
        :return: The vector.
        """
        return Vec2(0, 1)

    @staticmethod
    def down():
        """
        Returns a vector pointing down.
        :return: The vector.
        """
        return Vec2(0, -1)

    @staticmethod
    def right():
        """
        Returns a vector pointing right.
        :return: The vector.
        """
        return Vec2(1, 0)

    @staticmethod
    def left():
        """
        Returns a vector pointing left.
        :return: The vector.
        """
        return Vec2(-1, 0)

    @staticmethod
    def forward():
        """
        Returns a vector pointing forward.
        :return: The vector.
        """
        return Vec2(0, 1)

    @staticmethod
    def backward():
        """
        Returns a vector pointing backward.
        :return: The vector.
        """
        return Vec2(0, -1)

    @staticmethod
    def max(a: "Vec2", b: "Vec2"):
        """
        Returns the maximum of two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The maximum vector.
        """
        return Vec2(max(a.x, b.x), max(a.y, b.y))

    @staticmethod
    def min(a: "Vec2", b: "Vec2"):
        """
        Returns the minimum of two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The minimum vector.
        """
        return Vec2(min(a.x, b.x), min(a.y, b.y))

    def lerp(self, other: "Vec2", t):
        """
        Returns a vector between two vectors.
        :param other: The other vector.
        :param t: The interpolation value.
        :return: The interpolated vector.
        """
        return Vec2(self.x + (other.x - self.x) * t, self.y + (other.y - self.y) * t)

    def __add__(self, other):
        return Vec2(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Vec2(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        return Vec2(self.x * other, self.y * other)

    def __truediv__(self, other):
        return Vec2(self.x / other, self.y / other)

    def __floordiv__(self, other):
        return Vec2(self.x // other, self.y // other)

    def __mod__(self, other):
        return Vec2(self.x % other, self.y % other)

    def __str__(self):
        return "({}, {})".format(self.x, self.y)

    def __repr__(self):
        return "({}, {})".format(self.x, self.y)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y

    def __gt__(self, other):
        return self.x > other.x and self.y > other.y

    def __ge__(self, other):
        return self.x >= other.x and self.y >= other.y

    def __lt__(self, other):
        return self.x < other.x and self.y < other.y

    def __le__(self, other):
        return self.x <= other.x and self.y <= other.y

    def __abs__(self):
        return Vec2(abs(self.x), abs(self.y))

    def __neg__(self):
        return Vec2(-self.x, -self.y)

    def __round__(self):
        return Vec2(round(self.x), round(self.y))

    def __floor__(self):
        return Vec2(floor(self.x), floor(self.y))

    def __ceil__(self):
        return Vec2(ceil(self.x), ceil(self.y))

    def __trunc__(self):
        return Vec2(trunc(self.x), trunc(self.y))

    def __index__(self):
        return self.x, self.y

    def __getitem__(self, index):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        else:
            raise IndexError("Index out of range")

    def __setitem__(self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        else:
            raise IndexError("Index out of range")

    def __len__(self):
        return 2

    def __iter__(self):
        yield self.x
        yield self.y

    def __contains__(self, item):
        return item in (self.x, self.y)

    def __hash__(self):
        return hash((self.x, self.y))

    def __bool__(self):
        return self.x != 0 or self.y != 0

    def __nonzero__(self):
        return self.x != 0 or self.y != 0


# Vector3
class Vec3:
    def __init__(self, x=0, y=0, z=0):
        self.x = x
        self.y = y
        self.z = z

    def __int__(self, other: "Vec3"):
        self.x = other.x
        self.y = other.y
        self.z = other.z

    def distance(self, other: "Vec3"):
        """
        Returns the distance between two vectors.
        :param other: The other vector.
        :return: The distance.
        """
        return abs(self - other)

    def scale(self, other):
        """
        Returns the scaled vector.
        :param other: The scale value.
        :return: The scaled vector.
        """
        return Vec3(self.x * other, self.y * other, self.z * other)

    def reflect(self, other: "Vec3"):
        """
        Returns the reflected vector.
        :param other: The normal vector.
        :return: The reflected vector.
        """
        return self - other * 2 * self.dot(other)

    def angle_between(self, other: "Vec3"):
        """
        Returns the angle between two vectors.
        :param other: The other vector.
        :return: The angle.
        """
        return acos(self.dot(other) / (self.length() * other.length()))

    def normalize(self):
        """
        Returns the normalized vector.
        :return: The normalized vector.
        """
        return self / self.length()

    def dot(self, other: "Vec3"):
        """
        Returns the dot product of two vectors.
        :param other: The other vector.
        :return: The dot product.
        """
        return self.x * other.x + self.y * other.y + self.z * other.z

    def length(self):
        """
        Returns the length of the vector.
        :return: The length.
        """
        return (self.x ** 2 + self.y ** 2 + self.z ** 2) ** 0.5

    def length_sqr(self):
        """
        Returns the length squared of the vector.
        :return: The length squared.
        """
        return self.x ** 2 + self.y ** 2 + self.z ** 2

    def lerp(self, other: "Vec3", t):
        """
        Returns the linear interpolated vector.
        :param other: The other vector.
        :param t: The interpolation value.
        :return: The linear interpolated vector.
        """
        return self + (other - self) * t

    @staticmethod
    def cross(a, b):
        """
        Returns the cross product of two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The cross product.
        """
        return Vec3(a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x)

    @staticmethod
    def up():
        """
        Returns the up vector.
        :return: The up vector.
        """
        return Vec3(0, 1, 0)

    @staticmethod
    def down():
        """
        Returns the down vector.
        :return: The down vector.
        """
        return Vec3(0, -1, 0)

    @staticmethod
    def left():
        """
        Returns the left vector.
        :return: The left vector.
        """
        return Vec3(-1, 0, 0)

    @staticmethod
    def right():
        """
        Returns the right vector.
        :return: The right vector.
        """
        return Vec3(1, 0, 0)

    @staticmethod
    def forward():
        """
        Returns the forward vector.
        :return: The forward vector.
        """
        return Vec3(0, 0, 1)

    @staticmethod
    def backward():
        """
        Returns the backward vector.
        :return: The backward vector.
        """
        return Vec3(0, 0, -1)

    @staticmethod
    def zero():
        """
        Returns the zero vector.
        :return: The zero vector.
        """
        return Vec3(0, 0, 0)

    @staticmethod
    def angle(a: "Vec3", b: "Vec3"):
        """
        Returns the angle between two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The angle.
        """
        return acos(a.dot(b) / (a.length() * b.length()))

    @staticmethod
    def max(a: "Vec3", b: "Vec3"):
        """
        Returns the maximum of two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The maximum vector.
        """
        return Vec3(max(a.x, b.x), max(a.y, b.y), max(a.z, b.z))

    @staticmethod
    def min(a: "Vec3", b: "Vec3"):
        """
        Returns the minimum of two vectors.
        :param a: The first vector.
        :param b: The second vector.
        :return: The minimum vector.
        """
        return Vec3(min(a.x, b.x), min(a.y, b.y), min(a.z, b.z))

    def __add__(self, other):
        return Vec3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Vec3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other):
        return Vec3(self.x * other, self.y * other, self.z * other)

    def __truediv__(self, other):
        return Vec3(self.x / other, self.y / other, self.z / other)

    def __floordiv__(self, other):
        return Vec3(self.x // other, self.y // other, self.z // other)

    def __mod__(self, other):
        return Vec3(self.x % other, self.y % other, self.z % other)

    def __str__(self):
        return "({}, {}, {})".format(self.x, self.y, self.z)

    def __repr__(self):
        return "({}, {}, {})".format(self.x, self.y, self.z)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y or self.z != other.z

    def __gt__(self, other):
        return self.x > other.x and self.y > other.y and self.z > other.z

    def __ge__(self, other):
        return self.x >= other.x and self.y >= other.y and self.z >= other.z

    def __lt__(self, other):
        return self.x < other.x and self.y < other.y and self.z < other.z

    def __le__(self, other):
        return self.x <= other.x and self.y <= other.y and self.z <= other.z

    def __abs__(self):
        return Vec3(abs(self.x), abs(self.y), abs(self.z))

    def __neg__(self):
        return Vec3(-self.x, -self.y, -self.z)

    def __round__(self):
        return Vec3(round(self.x), round(self.y), round(self.z))

    def __floor__(self):
        return Vec3(floor(self.x), floor(self.y), floor(self.z))

    def __ceil__(self):
        return Vec3(ceil(self.x), ceil(self.y), ceil(self.z))

    def __trunc__(self):
        return Vec3(trunc(self.x), trunc(self.y), trunc(self.z))

    def __index__(self):
        return self.x, self.y, self.z

    def __getitem__(self, index):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise IndexError("Index out of range")

    def __setitem__(self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.z = value
        else:
            raise IndexError("Index out of range")

    def __len__(self):
        return 3

    def __iter__(self):
        yield self.x
        yield self.y
        yield self.z

    def __contains__(self, item):
        return item in (self.x, self.y, self.z)

    def __hash__(self):
        return hash((self.x, self.y, self.z))

    def __bool__(self):
        return self.x != 0 or self.y != 0 or self.z != 0

    def __nonzero__(self):
        return self.x != 0 or self.y != 0 or self.z != 0



from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

# setup build python setup.py sdist bdist_wheel
# upload twine upload dist/*
#build doc
# pdoc --html pyopt_tools -o docs


VERSION = '0.4'
DESCRIPTION = 'A Python package full of small & useful functions to make your life easier.'

setup(
    name="pyopt-tools",
    version=VERSION,
    author="FUSEN",
    author_email="fus3ngames@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    long_description=long_description,
    install_requires=[],
    keywords=['python', 'pyopt-tools', 'python optimization tools', 'python mirco functions', 'string tools',
              'math_tools', 'list_tool'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)
